"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.isCons = exports.isNil = exports.apS = exports.bind = exports.bindTo = exports.list = exports.Traversable = exports.Foldable = exports.Monad = exports.Applicative = exports.Apply = exports.Functor = exports.getMonoid = exports.getSemigroup = exports.getShow = exports.getEq = exports.URI = exports.findIndex = exports.of = exports.sequence = exports.foldMap = exports.reduceRight = exports.reduce = exports.chainFirst = exports.chain = exports.apSecond = exports.apFirst = exports.ap = exports.map = exports.dropLeftWhile = exports.dropLeft = exports.reverse = exports.toReversedArray = exports.toArray = exports.foldLeft = exports.tail = exports.head = exports.fromArray = exports.cons = exports.nil = void 0;
var A = __importStar(require("fp-ts/lib/Array"));
var Foldable_1 = require("fp-ts/lib/Foldable");
var function_1 = require("fp-ts/lib/function");
var Monoid_1 = require("fp-ts/lib/Monoid");
var O = __importStar(require("fp-ts/lib/Option"));
var pipeable_1 = require("fp-ts/lib/pipeable");
// -------------------------------------------------------------------------------------
// constructors
// -------------------------------------------------------------------------------------
/**
 * @category constructors
 * @since 0.1.8
 */
exports.nil = { type: 'Nil', length: 0 };
/**
 * Attaches an element to the front of a list.
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.deepStrictEqual(L.cons('a', L.nil), { type: 'Cons', head: 'a', tail: L.nil, length: 1 })
 *
 * @category constructors
 * @since 0.1.8
 */
var cons = function (head, tail) { return ({
    type: 'Cons',
    head: head,
    tail: tail,
    length: 1 + tail.length
}); };
exports.cons = cons;
/**
 * Creates a list from an array
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.deepStrictEqual(L.fromArray([]), L.nil)
 * assert.deepStrictEqual(L.fromArray(['a', 'b']), L.cons('a', L.of('b')))
 *
 * @category constructors
 * @since 0.1.8
 */
var fromArray = function (as) { return A.array.reduceRight(as, exports.nil, exports.cons); };
exports.fromArray = fromArray;
// -------------------------------------------------------------------------------------
// destructors
// -------------------------------------------------------------------------------------
/**
 * Gets the first element in a list, or `None` if the list is empty.
 *
 * @example
 * import * as O from 'fp-ts/Option'
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.deepStrictEqual(L.head(L.nil), O.none)
 * assert.deepStrictEqual(L.head(L.cons('x', L.of('a'))), O.some('x'))
 *
 * @category destructors
 * @since 0.1.8
 */
var head = function (fa) { return ((0, exports.isCons)(fa) ? O.some(fa.head) : O.none); };
exports.head = head;
/**
 * Gets all but the first element of a list, or `None` if the list is empty.
 *
 * @example
 * import * as O from 'fp-ts/Option'
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.deepStrictEqual(L.tail(L.nil), O.none)
 * assert.deepStrictEqual(L.tail(L.of('a')), O.some(L.nil))
 * assert.deepStrictEqual(L.tail(L.cons('x', L.of('a'))), O.some(L.of('a')))
 *
 * @category destructors
 * @since 0.1.8
 */
var tail = function (fa) { return ((0, exports.isCons)(fa) ? O.some(fa.tail) : O.none); };
exports.tail = tail;
/**
 * Breaks a list into its first element and the remaining elements.
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * const len: <A>(as: L.List<A>) => number = L.foldLeft(
 *   () => 0,
 *   (_, tail) => 1 + len(tail)
 * )
 * assert.deepStrictEqual(len(L.cons('a', L.of('b'))), 2)
 *
 * @category destructors
 * @since 0.1.8
 */
var foldLeft = function (onNil, onCons) { return function (fa) { return ((0, exports.isNil)(fa) ? onNil() : onCons(fa.head, fa.tail)); }; };
exports.foldLeft = foldLeft;
/**
 * Gets an array from a list.
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.deepStrictEqual(L.toArray(L.cons('a', L.of('b'))), ['a', 'b'])
 *
 * @category destructors
 * @since 0.1.8
 */
var toArray = function (fa) {
    var length = fa.length;
    var out = new Array(length);
    var l = fa;
    for (var i = 0; i < length; i++) {
        out[i] = l.head;
        l = l.tail;
    }
    return out;
};
exports.toArray = toArray;
/**
 * Gets an array from a list in a reversed order.
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.deepStrictEqual(L.toReversedArray(L.cons('a', L.of('b'))), ['b', 'a'])
 *
 * @category destructors
 * @since 0.1.8
 */
var toReversedArray = function (fa) {
    var length = fa.length;
    var out = new Array(length);
    var l = fa;
    for (var i = 0; i < length; i++) {
        out[length - i - 1] = l.head;
        l = l.tail;
    }
    return out;
};
exports.toReversedArray = toReversedArray;
// -------------------------------------------------------------------------------------
// combinators
// -------------------------------------------------------------------------------------
/**
 * Reverse a list.
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.deepStrictEqual(L.reverse(L.cons(1, L.cons(2, L.of(3)))), L.cons(3, L.cons(2, L.of(1))))
 *
 * @category combinators
 * @since 0.1.8
 */
var reverse = function (fa) {
    var out = exports.nil;
    var l = fa;
    while ((0, exports.isCons)(l)) {
        out = (0, exports.cons)(l.head, out);
        l = l.tail;
    }
    return out;
};
exports.reverse = reverse;
/**
 * Drops the specified number of elements from the front of a list.
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.deepStrictEqual(L.dropLeft(1)(L.nil), L.nil)
 * assert.deepStrictEqual(L.dropLeft(1)(L.cons(1, L.of(2))), L.of(2))
 * assert.deepStrictEqual(L.dropLeft(3)(L.cons(1, L.of(2))), L.nil)
 *
 * @category combinators
 * @since 0.1.8
 */
var dropLeft = function (n) { return function (fa) {
    if ((0, exports.isNil)(fa))
        return exports.nil;
    var i = 0;
    var l = fa;
    while ((0, exports.isCons)(l) && i < n) {
        i++;
        l = l.tail;
    }
    return l;
}; };
exports.dropLeft = dropLeft;
function dropLeftWhile(predicate) {
    return function (fa) {
        if ((0, exports.isNil)(fa))
            return exports.nil;
        var l = fa;
        while ((0, exports.isCons)(l) && predicate(l.head)) {
            l = l.tail;
        }
        return l;
    };
}
exports.dropLeftWhile = dropLeftWhile;
// -------------------------------------------------------------------------------------
// non-pipeables
// -------------------------------------------------------------------------------------
var map_ = function (fa, f) { return (0, pipeable_1.pipe)(fa, (0, exports.map)(f)); };
var ap_ = function (fab, fa) { return (0, pipeable_1.pipe)(fab, (0, exports.ap)(fa)); };
var chain_ = function (ma, f) { return (0, pipeable_1.pipe)(ma, (0, exports.chain)(f)); };
var reduce_ = function (fa, b, f) { return (0, pipeable_1.pipe)(fa, (0, exports.reduce)(b, f)); };
var reduceRight_ = function (fa, b, f) { return (0, pipeable_1.pipe)(fa, (0, exports.reduceRight)(b, f)); };
var foldMap_ = function (M) { return function (fa, f) { return (0, pipeable_1.pipe)(fa, (0, exports.foldMap)(M)(f)); }; };
var traverse_ = function (F) {
    return function (ta, f) {
        return exports.list.reduceRight(ta, F.of(exports.nil), function (a, fbs) {
            return F.ap(F.map(fbs, function (bs) { return function (b) { return (0, exports.cons)(b, bs); }; }), f(a));
        });
    };
};
var sequence_ = function (F) { return function (ta) {
    return exports.list.reduceRight(ta, F.of(exports.nil), function (a, fas) {
        return F.ap(F.map(fas, function (as) { return function (a) { return (0, exports.cons)(a, as); }; }), a);
    });
}; };
// -------------------------------------------------------------------------------------
// pipeables
// -------------------------------------------------------------------------------------
/**
 * @category Functor
 * @since 0.1.18
 */
var map = function (f) { return function (fa) {
    return (0, pipeable_1.pipe)((0, exports.toArray)(fa), A.reduceRight(exports.nil, function (a, b) { return (0, exports.cons)(f(a), b); }));
}; };
exports.map = map;
/**
 * @category Functor
 * @since 0.1.20
 */
var ap = function (fa) { return (0, exports.chain)(function (f) { return (0, pipeable_1.pipe)(fa, (0, exports.map)(f)); }); };
exports.ap = ap;
/**
 * @category Apply
 * @since 0.1.20
 */
var apFirst = function (fb) {
    return (0, function_1.flow)((0, exports.map)(function (a) { return function () { return a; }; }), (0, exports.ap)(fb));
};
exports.apFirst = apFirst;
/**
 * @category Apply
 * @since 0.1.20
 */
var apSecond = function (fb) {
    return (0, function_1.flow)((0, exports.map)(function () { return function (b) { return b; }; }), (0, exports.ap)(fb));
};
exports.apSecond = apSecond;
/**
 * @category Monad
 * @since 0.1.20
 */
var chain = function (f) { return function (ma) {
    return (0, pipeable_1.pipe)(ma, (0, exports.foldLeft)(function () { return exports.nil; }, function (x, xs) {
        var S = (0, exports.getSemigroup)();
        var out = f(x);
        var l = xs;
        while ((0, exports.isCons)(l)) {
            out = S.concat(out, f(l.head));
            l = l.tail;
        }
        return out;
    }));
}; };
exports.chain = chain;
/**
 * @category Monad
 * @since 0.1.20
 */
var chainFirst = function (f) {
    return (0, exports.chain)(function (a) {
        return (0, pipeable_1.pipe)(f(a), (0, exports.map)(function () { return a; }));
    });
};
exports.chainFirst = chainFirst;
/**
 * @category Foldable
 * @since 0.1.18
 */
var reduce = function (b, f) { return function (fa) {
    var out = b;
    var l = fa;
    while ((0, exports.isCons)(l)) {
        out = f(out, l.head);
        l = l.tail;
    }
    return out;
}; };
exports.reduce = reduce;
/**
 * @category Foldable
 * @since 0.1.18
 */
var reduceRight = function (b, f) { return function (fa) {
    return (0, pipeable_1.pipe)((0, exports.toArray)(fa), A.reduceRight(b, f));
}; };
exports.reduceRight = reduceRight;
/**
 * @category Foldable
 * @since 0.1.18
 */
var foldMap = function (M) { return function (f) { return function (fa) {
    var out = M.empty;
    var l = fa;
    while ((0, exports.isCons)(l)) {
        out = M.concat(out, f(l.head));
        l = l.tail;
    }
    return out;
}; }; };
exports.foldMap = foldMap;
// TODO: add pipeable traverse when fp-ts version >= 2.6.3
// /**
//  * @category Traversable
//  * @since 0.1.18
//  */
// export const traverse: PipeableTraverse1<URI> = <F>(
//   F: Applicative<F>
// ): (<A, B>(f: (a: A) => HKT<F, B>) => (ta: List<A>) => HKT<F, List<B>>) => {
//   const traverseF = traverse_(F)
//   return f => ta => traverseF(ta, f)
// }
/**
 * @category Traversable
 * @since 0.1.18
 */
var sequence = function (F) { return sequence_(F); };
exports.sequence = sequence;
/**
 * Creates a list with a single element.
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.deepStrictEqual(L.of('a'), L.cons('a', L.nil))
 *
 * @category Applicative
 * @since 0.1.8
 */
var of = function (head) { return (0, exports.cons)(head, exports.nil); };
exports.of = of;
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * Finds the first index for which a predicate holds.
 *
 * @example
 * import * as O from 'fp-ts/Option'
 * import * as L from 'fp-ts-contrib/List'
 *
 * const f = (a: number): boolean => a % 2 === 0
 * const findIndexEven = L.findIndex(f)
 * assert.deepStrictEqual(findIndexEven(L.nil), O.none)
 * assert.deepStrictEqual(findIndexEven(L.cons(1, L.of(2))), O.some(1))
 * assert.deepStrictEqual(findIndexEven(L.of(1)), O.none)
 *
 * @since 0.1.8
 */
var findIndex = function (predicate) { return function (fa) {
    var l = fa;
    var i = 0;
    while ((0, exports.isCons)(l)) {
        if (predicate(l.head))
            return O.some(i);
        l = l.tail;
        i++;
    }
    return O.none;
}; };
exports.findIndex = findIndex;
// -------------------------------------------------------------------------------------
// instances
// -------------------------------------------------------------------------------------
/**
 * @category instances
 * @since 0.1.8
 */
exports.URI = 'List';
/**
 * Derives an `Eq` over the `List` of a given element type from the `Eq` of that type.
 * The derived `Eq` defines two lists as equal if all elements of both lists
 * are compared equal pairwise with the given `E`. In case of lists of different
 * lengths, the result is non equality.
 *
 * @example
 * import { eqString } from 'fp-ts/Eq'
 * import * as L from 'fp-ts-contrib/List'
 *
 * const E = L.getEq(eqString)
 * assert.strictEqual(E.equals(L.cons('a', L.of('b')), L.cons('a', L.of('b'))), true)
 * assert.strictEqual(E.equals(L.of('x'), L.nil), false)
 *
 * @category instances
 * @since 0.1.8
 */
var getEq = function (E) { return ({
    equals: function (x, y) {
        if (x.length !== y.length)
            return false;
        var lx = x;
        var ly = y;
        while ((0, exports.isCons)(lx) && (0, exports.isCons)(ly)) {
            if (!E.equals(lx.head, ly.head))
                return false;
            lx = lx.tail;
            ly = ly.tail;
        }
        return true;
    }
}); };
exports.getEq = getEq;
/**
 * @category instances
 * @since 0.1.20
 */
var getShow = function (S) { return ({
    show: function (as) {
        return (0, pipeable_1.pipe)(as, (0, exports.map)(S.show), (0, exports.foldLeft)(function () { return 'Nil'; }, function (x, xs) { return "(".concat((0, Foldable_1.intercalate)(Monoid_1.monoidString, exports.Foldable)(' : ', (0, exports.cons)(x, xs)), " : Nil)"); }));
    }
}); };
exports.getShow = getShow;
/**
 * @category instances
 * @since 0.1.20
 */
var getSemigroup = function () { return ({
    concat: function (xs, ys) { return (0, pipeable_1.pipe)(xs, (0, exports.reduceRight)(ys, exports.cons)); }
}); };
exports.getSemigroup = getSemigroup;
/**
 * @category instances
 * @since 0.1.20
 */
var getMonoid = function () { return (__assign(__assign({}, (0, exports.getSemigroup)()), { empty: exports.nil })); };
exports.getMonoid = getMonoid;
/**
 * @category instances
 * @since 0.1.18
 */
exports.Functor = {
    URI: exports.URI,
    map: map_
};
/**
 * @category instances
 * @since 0.1.20
 */
exports.Apply = {
    URI: exports.URI,
    map: map_,
    ap: ap_
};
/**
 * @category instances
 * @since 0.1.20
 */
exports.Applicative = {
    URI: exports.URI,
    map: map_,
    ap: ap_,
    of: exports.of
};
/**
 * @category instances
 * @since 0.1.20
 */
exports.Monad = {
    URI: exports.URI,
    map: map_,
    ap: ap_,
    of: exports.of,
    chain: chain_
};
/**
 * @category instances
 * @since 0.1.18
 */
exports.Foldable = {
    URI: exports.URI,
    foldMap: foldMap_,
    reduce: reduce_,
    reduceRight: reduceRight_
};
/**
 * @category instances
 * @since 0.1.18
 */
exports.Traversable = {
    URI: exports.URI,
    map: map_,
    foldMap: foldMap_,
    reduce: reduce_,
    reduceRight: reduceRight_,
    traverse: traverse_,
    sequence: sequence_
};
/**
 * @category instances
 * @since 0.1.8
 */
exports.list = {
    URI: exports.URI,
    map: map_,
    reduce: reduce_,
    foldMap: foldMap_,
    reduceRight: reduceRight_,
    traverse: traverse_,
    sequence: sequence_
};
// -------------------------------------------------------------------------------------
// do notation
// -------------------------------------------------------------------------------------
/**
 * @internal
 */
var bind_ = function (a, name, b) {
    var _a;
    return Object.assign({}, a, (_a = {}, _a[name] = b, _a));
};
/**
 * @since 0.1.20
 */
var bindTo = function (name) { return function (fa) {
    return (0, pipeable_1.pipe)(fa, (0, exports.map)(function (a) { return bind_({}, name, a); }));
}; };
exports.bindTo = bindTo;
/**
 * @since 0.1.20
 */
var bind = function (name, f) { return function (fa) {
    return (0, pipeable_1.pipe)(fa, (0, exports.chain)(function (a) {
        return (0, pipeable_1.pipe)(f(a), (0, exports.map)(function (b) { return bind_(a, name, b); }));
    }));
}; };
exports.bind = bind;
// -------------------------------------------------------------------------------------
// pipeable sequence S
// -------------------------------------------------------------------------------------
/**
 * @since 0.1.20
 */
var apS = function (name, fb) {
    return (0, function_1.flow)((0, exports.map)(function (a) { return function (b) { return bind_(a, name, b); }; }), (0, exports.ap)(fb));
};
exports.apS = apS;
// -------------------------------------------------------------------------------------
// utils
// -------------------------------------------------------------------------------------
/**
 * Tests whether a list is an empty list.
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.strictEqual(L.isNil(L.nil), true)
 * assert.strictEqual(L.isNil(L.of(6)), false)
 *
 * @since 0.1.8
 */
var isNil = function (a) { return a.type === 'Nil'; };
exports.isNil = isNil;
/**
 * Tests whether a list is a non empty list.
 *
 * @example
 * import * as L from 'fp-ts-contrib/List'
 *
 * assert.strictEqual(L.isCons(L.nil), false)
 * assert.strictEqual(L.isCons(L.of(1)), true)
 *
 * @since 0.1.8
 */
var isCons = function (a) { return a.type === 'Cons'; };
exports.isCons = isCons;
